import unittest
from z85base91 import B91


class TestB91(unittest.TestCase):
    def test_encode_empty(self):
        """Test encoding an empty byte sequence."""
        self.assertEqual(B91.encode(b''), b'')
        self.assertEqual(B91.encode(''), b'')

    def test_decode_empty(self):
        """Test decoding an empty string."""
        self.assertEqual(B91.decode(''), b'')
        self.assertEqual(B91.decode(b''), b'')

    def test_encode_single_byte(self):
        """Test encoding a single byte."""
        self.assertEqual(b'A', B91.decode(B91.encode(b'A')))
        self.assertEqual(b'B', B91.decode(B91.encode('B')))
        self.assertEqual(b'_~', B91.decode(B91.encode(b'_~')))
        self.assertEqual(b'_~', B91.decode(B91.encode('_~')))

    def test_encode_short_string(self):
        """Test encoding a short string."""
        self.assertEqual(b'hello', B91.decode(B91.encode(b'hello')))
        self.assertEqual(B91.decode('>OwJh>Io0Tv!lE'), b'Hello World')

    def test_encode_decode_round_trip(self):
        """Test encoding and decoding round-trip."""
        data = b'The quick brown fox jumps over the lazy dog.'
        encoded = B91.encode(data)
        decoded = B91.decode(encoded)
        self.assertEqual(decoded, data)

    def test_encode_unicode_string(self):
        """Test encoding a Unicode string."""
        data = 'こんにちは'  # Japanese for "hello"
        encoded = B91.encode(data)
        decoded = B91.decode(encoded)
        self.assertEqual(decoded.decode('utf-8'), data)

    def test_decode_invalid_character(self):
        """Test decoding with invalid Base91 characters."""
        with self.assertRaises(ValueError):
            B91.decode('Invalid🎉Chars')

    def test_3bytes_threshold(self):
        """Test edge cases around the 88 threshold."""
        data = b'\x00\x00\x00'  # Minimal data
        encoded = B91.encode(data)
        self.assertEqual(B91.decode(encoded), data)

    def test_encode_large_data(self):
        """Test encoding a large byte sequence."""
        data = b'\xff' * 1000
        encoded = B91.encode(data)
        decoded = B91.decode(encoded)
        self.assertEqual(decoded, data)


if __name__ == '__main__':
    unittest.main()



import base64
import pybase64
import random
import string
import time
from typing import Callable, Dict, List, Tuple

import click
from z85base91 import Z85P, Z85B, B91
from hivemind_bus_client.encodings import Z85B as Z85Bpy, Z85P as Z85Ppy, B91 as B91py
from tabulate import tabulate


def get_encoder(encoding: str) -> Callable[[bytes], bytes]:
    """Retrieve the encoder function for the given encoding."""
    encoders = {
        "base64": base64.b64encode,
        "base64_py": pybase64.b64encode,
        "z85b": Z85B.encode,
        "z85p": Z85P.encode,
        "base91": B91.encode,
        "z85b_py": Z85Bpy.encode,
        "z85p_py": Z85Ppy.encode,
        "base91_py": B91py.encode,
        "base32": base64.b32encode
    }
    return encoders[encoding]


def get_decoder(encoding: str) -> Callable[[bytes], bytes]:
    """Retrieve the decoder function for the given encoding."""
    decoders = {
        "base64": base64.b64decode,
        "base64_py": pybase64.b64decode,
        "z85b": Z85B.decode,
        "z85p": Z85P.decode,
        "base91": B91.decode,
        "z85b_py": Z85Bpy.decode,
        "z85p_py": Z85Ppy.decode,
        "base91_py": B91py.decode,
        "base32": base64.b32decode
    }
    return decoders[encoding]


def generate_random_data(size: int) -> bytes:
    """Generate random binary data of a given size."""
    return ''.join(random.choices(string.ascii_letters + string.digits, k=size)).encode("utf-8")


def benchmark_encoding(encoding: str, data: bytes) -> Dict[str, int]:
    """Benchmark encoding and decoding for a given encoding."""
    encoder = get_encoder(encoding)
    decoder = get_decoder(encoding)

    # Measure encoding time in nanoseconds
    start_time = time.perf_counter_ns()
    encoded_data = encoder(data)
    encoding_time = time.perf_counter_ns() - start_time

    # Measure decoding time in nanoseconds
    start_time = time.perf_counter_ns()
    decoded_data = decoder(encoded_data)
    decoding_time = time.perf_counter_ns() - start_time

    # Validate decoding
    if decoded_data != data:
        raise ValueError(f"Decoded data does not match for encoding {encoding}.")

    # Calculate size increase
    original_size = len(data)
    encoded_size = len(encoded_data)
    size_increase = encoded_size / original_size

    return {
        "encoding_time": encoding_time,
        "decoding_time": decoding_time,
        "size_increase": size_increase,
    }


def get_rankings(metric: Dict[str, Dict[str, int]], key: str) -> List[Tuple[str, int]]:
    """Rank the encodings based on the provided metric, handling ties."""
    sorted_encodings = sorted(metric.items(), key=lambda x: x[1][key], reverse=False)
    rankings = []
    current_rank = 1  # Start from rank 1

    for i in range(len(sorted_encodings)):
        if i > 0 and sorted_encodings[i][1][key] == sorted_encodings[i - 1][1][key]:
            # Tie case: Same rank as the previous item
            rankings.append((sorted_encodings[i][0], rankings[-1][1]))
        else:
            # No tie, increase the rank
            rankings.append((sorted_encodings[i][0], current_rank))
            current_rank += 1  # Increment rank only when no tie

    return rankings


def compare_python_c(encoding: str, python_results: Dict[str, Dict[str, int]],
                     c_results: Dict[str, Dict[str, int]]) -> float:
    """Compare the speed between Python and C encodings and calculate how many times faster or slower it is."""
    if encoding.endswith("_py"):
        c_encoding = encoding[:-3]  # Remove the _py suffix to get the C counterpart
        python_time = python_results[encoding]["encoding_time"]
        c_time = c_results[c_encoding]["encoding_time"]

        if c_time == 0:
            return float('inf')  # Avoid division by zero
        return python_time / c_time
    return 1.0  # If not a Python version, return 1 (no comparison)


@click.command()
@click.option("--sizes", default="100,1000,10000", help="Comma-separated list of data sizes to test.")
@click.option("--iterations", default=10, help="Number of iterations for each test.")
def main(sizes: str, iterations: int):
    sizes = list(map(int, sizes.split(",")))

    encodings = [
        "base64",
        "base32",
        "base64_py",
        "z85b",
        "z85p",
        "base91",
        "z85b_py",
        "z85p_py",
        "base91_py",
    ]

    results = {size: {encoding: [] for encoding in encodings} for size in sizes}

    # Run benchmarks
    for size in sizes:
        print(f"Testing size: {size} bytes")
        for _ in range(iterations):
            data = generate_random_data(size)
            for encoding in encodings:
                result = benchmark_encoding(encoding, data)
                results[size][encoding].append(result)

    # Calculate averages and print results
    global_ranking = {encoding: {"encoding_time": 0, "decoding_time": 0, "size_increase": 0} for encoding in encodings}
    table = []

    # Aggregate results across all sizes
    for encoding in encodings:
        total_encoding_time = 0
        total_decoding_time = 0
        total_size_increase = 0

        for size, encoding_results in results.items():
            avg_encoding_time = sum(m["encoding_time"] for m in encoding_results[encoding]) // iterations
            avg_decoding_time = sum(m["decoding_time"] for m in encoding_results[encoding]) // iterations
            avg_size_increase = sum(m["size_increase"] for m in encoding_results[encoding]) / iterations

            total_encoding_time += avg_encoding_time
            total_decoding_time += avg_decoding_time
            total_size_increase += avg_size_increase

            table.append([
                encoding,
                f"{avg_encoding_time} ns",
                f"{avg_decoding_time} ns",
                f"{avg_size_increase:.2f}x size increase"
            ])

        # Store global averages
        global_ranking[encoding]["encoding_time"] = total_encoding_time // len(sizes)
        global_ranking[encoding]["decoding_time"] = total_decoding_time // len(sizes)
        global_ranking[encoding]["size_increase"] = total_size_increase / len(sizes)

    # Global ranking (based on average times)
    print("\n### Global Ranking (Merged) ###")

    # Get rankings for encoding time, decoding time, and size increase
    sorted_by_encoding_time = get_rankings(global_ranking, "encoding_time")
    sorted_by_decoding_time = get_rankings(global_ranking, "decoding_time")
    sorted_by_size_increase = get_rankings(global_ranking, "size_increase")

    merged_table = []
    for encoding, metrics in global_ranking.items():
        encoding_time_rank = next(rank for enc, rank in sorted_by_encoding_time if enc == encoding)
        decoding_time_rank = next(rank for enc, rank in sorted_by_decoding_time if enc == encoding)
        size_increase_rank = next(rank for enc, rank in sorted_by_size_increase if enc == encoding)

        # Calculate the average rank
        avg_rank = (encoding_time_rank + decoding_time_rank + size_increase_rank) / 3

        # Medal assignments
        encoding_time_medal = "🥇" if encoding_time_rank == 1 else "🥈" if encoding_time_rank == 2 else "🥉" if encoding_time_rank == 3 else ""
        decoding_time_medal = "🥇" if decoding_time_rank == 1 else "🥈" if decoding_time_rank == 2 else "🥉" if decoding_time_rank == 3 else ""
        size_increase_medal = "🥇" if size_increase_rank == 1 else "🥈" if size_increase_rank == 2 else "🥉" if size_increase_rank == 3 else ""

        # Add the top-ranked emojis
        merged_table.append([
            encoding,
            f"{metrics['encoding_time']} ns",
            f"{metrics['decoding_time']} ns",
            f"{metrics['size_increase']:.2f}x",
            f"{encoding_time_rank} {encoding_time_medal}",
            f"{decoding_time_rank} {decoding_time_medal}",
            f"{size_increase_rank} {size_increase_medal}",
            avg_rank
        ])

        # Pairwise comparison for Python vs C
        if encoding.endswith("_py"):  # If it's a Python version
            speed_comparison = compare_python_c(encoding, global_ranking, global_ranking)
            if float(speed_comparison) > 1:
                merged_table[-1].append(f"{speed_comparison:.2f}x slower")
            else:
                merged_table[-1].append(f"{1/speed_comparison:.2f}x faster")

    # Sort the merged table based on the average rank
    merged_table.sort(key=lambda x: float(str(x[-2]).split()[0]), reverse=False)

    # Display the final table
    print(tabulate(merged_table,
                   headers=["Encoding", "Avg Encoding Time (ns)", "Avg Decoding Time (ns)", "Avg Size Increase",
                            "Encoding Rank", "Decoding Rank", "Size Increase Rank", "Score", "Reference vs Optimized"],
                   tablefmt="grid"))


if __name__ == "__main__":
    main()



import ctypes
import logging
import os.path
import platform
from ctypes import c_char_p, c_void_p, c_size_t, c_ubyte, byref, POINTER
from typing import Union

# Set up logging
logging.basicConfig(level=logging.WARNING, format='%(asctime)s - %(levelname)s - %(message)s')


# get the appropriate shared library file based on architecture
def get_arch_lib(lib_base_name: str) -> str:
    """
    Returns the path to the shared library based on the system's architecture.

    Args:
        lib_base_name (str): The base name of the shared library (e.g., 'libz85p').

    Returns:
        str: The path to the shared library for the appropriate system architecture.

    Raises:
        ValueError: If the system architecture is unsupported.
    """
    arch = platform.machine()  # Get system architecture
    if arch == "x86_64":
        return f"{os.path.dirname(__file__)}/{lib_base_name}-x86_64.so"
    elif arch == "aarch64":
        return f"{os.path.dirname(__file__)}/{lib_base_name}-aarch64.so"
    elif arch == "i386" or arch == "i686":
        return f"{os.path.dirname(__file__)}/{lib_base_name}-i386.so"
    else:
        raise ValueError(f"Unsupported architecture: {arch}")


try:
    class Z85P:
        """
        Class for encoding and decoding Z85P format using a C-based shared library.
        If the C library is not available, it falls back to a pure Python implementation.
        """
        Z85CHARS = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.-:+=^!/*?&<>()[]{}@%$#"

        # Load the correct shared library based on system architecture
        lib = ctypes.CDLL(get_arch_lib('libz85p'))

        # Initialize the Z85 map (this needs to be called first)
        lib.initialize_z85_map()

        # Define the encode function prototype
        lib.encode_z85p.argtypes = [ctypes.POINTER(ctypes.c_ubyte), c_size_t, POINTER(c_size_t)]
        lib.encode_z85p.restype = ctypes.POINTER(ctypes.c_ubyte)

        # Define the decode function prototype
        lib.decode_z85p.argtypes = [ctypes.POINTER(ctypes.c_ubyte), c_size_t, POINTER(c_size_t)]
        lib.decode_z85p.restype = ctypes.POINTER(ctypes.c_ubyte)

        @classmethod
        def encode(cls, data: Union[str, bytes]) -> bytes:
            """
            Encodes the input data into Z85P format.

            Args:
                data (bytes): The raw data to encode.

            Returns:
                bytes: The Z85P-encoded data.

            Raises:
                ValueError: If encoding fails.
            """
            if isinstance(data, str):
                data = data.encode('utf-8')

            out_len = c_size_t(0)
            raw_data = (ctypes.c_ubyte * len(data))(*data)
            encoded_data = cls.lib.encode_z85p(raw_data, len(data), ctypes.byref(out_len))
            return bytes(ctypes.string_at(encoded_data, out_len.value))

        @classmethod
        def decode(cls, encoded_data: Union[str, bytes]) -> bytes:
            """
            Decodes the input Z85P-encoded data into raw bytes.

            Args:
                encoded_data (bytes): The Z85P-encoded data to decode.

            Returns:
                bytes: The decoded raw data.

            Raises:
                ValueError: If decoding fails.
            """
            if isinstance(encoded_data, str):
                if any(c not in Z85P.Z85CHARS for c in encoded_data):
                    raise ValueError("Invalid Z85 character")
                encoded_data = encoded_data.encode('utf-8')

            out_len = c_size_t(0)
            raw_data = (ctypes.c_ubyte * len(encoded_data))(*encoded_data)
            decoded_data = cls.lib.decode_z85p(raw_data, len(encoded_data), ctypes.byref(out_len))
            return bytes(ctypes.string_at(decoded_data, out_len.value))
except Exception as e:
    logging.warning(f"Z85P C library not available: {e}. Falling back to pure Python implementation.")
    from z85base91.z85p import Z85P

try:
    class B91:
        """
        Class for encoding and decoding Base91 format using a C-based shared library.
        If the C library is not available, it falls back to a pure Python implementation.
        """
        # Load the correct shared library based on system architecture
        lib = ctypes.CDLL(get_arch_lib('libbase91'))

        # Initialize the decode table (this needs to be called first)
        lib.initialize_decode_table()

        # Define the decode function prototype
        lib.decode.argtypes = [c_char_p, ctypes.POINTER(c_size_t)]
        lib.decode.restype = c_void_p

        # Define the encode function prototype
        lib.encode.argtypes = [ctypes.POINTER(ctypes.c_ubyte), c_size_t, ctypes.POINTER(c_size_t)]
        lib.encode.restype = c_char_p

        @classmethod
        def encode(cls, data: Union[str, bytes]) -> bytes:
            """
            Encodes the input data into Base91 format.

            Args:
                data (Union[str, bytes]): The raw data to encode.

            Returns:
                bytes: The Base91-encoded data.

            Raises:
                ValueError: If encoding fails.
            """
            if isinstance(data, str):
                # Convert the data to bytes
                data = data.encode('utf-8')
            output_len = c_size_t(0)

            # Call the C function
            encoded_data = cls.lib.encode((ctypes.c_ubyte * len(data))(*data), len(data), ctypes.byref(output_len))

            return ctypes.string_at(encoded_data, output_len.value)

        @classmethod
        def decode(cls, encoded_data: Union[str, bytes]) -> bytes:
            """
            Decodes the input Base91-encoded data into raw bytes.

            Args:
                encoded_data (Union[str, bytes]): The Base91-encoded data to decode.

            Returns:
                bytes: The decoded raw data.

            Raises:
                ValueError: If decoding fails.
            """
            if isinstance(encoded_data, str):
                # Convert the encoded data to bytes
                encoded_data = encoded_data.encode('utf-8')
            output_len = c_size_t(0)

            # Call the C function
            decoded_data = cls.lib.decode(encoded_data, ctypes.byref(output_len))

            if not decoded_data:
                raise ValueError("Invalid Base91 string")
            return ctypes.string_at(decoded_data, output_len.value)

except Exception as e:
    logging.warning(f"Base91 C library not available: {e}. Falling back to pure Python implementation.")
    from z85base91.b91 import B91

try:
    class Z85B:
        """
        Class for encoding and decoding Z85B format using a C-based shared library.
        If the C library is not available, it falls back to a pure Python implementation.
        """
        # Load the correct shared library based on system architecture
        lib = ctypes.CDLL(get_arch_lib('libz85b'))

        # Define function prototypes
        lib.encode_z85b.argtypes = [POINTER(c_ubyte), c_size_t, POINTER(c_size_t)]
        lib.encode_z85b.restype = POINTER(c_ubyte)

        lib.decode_z85b.argtypes = [POINTER(c_ubyte), c_size_t, POINTER(c_size_t)]
        lib.decode_z85b.restype = POINTER(c_ubyte)

        lib.free.argtypes = [ctypes.c_void_p]  # Add free function for memory cleanup

        @classmethod
        def encode(cls, data: Union[str, bytes]) -> bytes:
            """
            Encodes the input data into Z85B format.

            Args:
                data (bytes): The raw data to encode.

            Returns:
                bytes: The Z85B-encoded data.

            Raises:
                ValueError: If encoding fails.
            """
            if isinstance(data, str):
                data = data.encode('utf-8')
            output_len = c_size_t(0)
            encoded_data = cls.lib.encode_z85b((c_ubyte * len(data))(*data), len(data), byref(output_len))
            if not encoded_data:
                raise ValueError("Encoding failed")

            try:
                return ctypes.string_at(encoded_data, output_len.value)
            finally:
                cls.lib.free(encoded_data)

        @classmethod
        def decode(cls, encoded_data: Union[str, bytes]) -> bytes:
            """
            Decodes the input Z85B-encoded data into raw bytes.

            Args:
                encoded_data (bytes): The Z85B-encoded data to decode.

            Returns:
                bytes: The decoded raw data.

            Raises:
                ValueError: If decoding fails.
            """
            if isinstance(encoded_data, str):
                # Convert the encoded data to bytes
                if any(c not in Z85P.Z85CHARS for c in encoded_data):
                    raise ValueError("Invalid Z85 character")
                encoded_data = encoded_data.encode('utf-8')
            output_len = c_size_t(0)
            decoded_data = cls.lib.decode_z85b((c_ubyte * len(encoded_data))(*encoded_data), len(encoded_data),
                                               byref(output_len))
            try:
                return ctypes.string_at(decoded_data, output_len.value)
            finally:
                cls.lib.free(decoded_data)
except Exception as e:
    logging.warning(f"Z85B C library not available: {e}. Falling back to pure Python implementation.")
    from z85base91.z85b import Z85B

if __name__ == "__main__":

    from z85base91.b91 import B91 as B91py
    from z85base91.z85b import Z85B as Z85Bpy
    from z85base91.z85p import Z85P as Z85Ppy


    def test_b91(s=b"Hello, Base91!"):
        # Example usage:
        try:
            pencoded = B91py.encode(s)
            print("Encoded py:", pencoded)
            pdecoded = B91py.decode(pencoded)
            print("Decoded py:", pdecoded)

            encoded = B91.encode(s)
            print("Encoded:", encoded)
            decoded = B91.decode(encoded)
            print("Decoded:", decoded)

            assert pdecoded == decoded
            assert pencoded == encoded
        except Exception as e:
            print(f"Error: {e}")


    def test_z85b(s=b"Hello, Z85B!"):
        try:
            pencoded = Z85Bpy.encode(s)
            print("Encoded py:", pencoded)
            pdecoded = Z85Bpy.decode(pencoded)
            print("Decoded py:", pdecoded)

            encoded = Z85B.encode(s)
            print("Encoded:", encoded)
            decoded = Z85B.decode(encoded)
            print("Decoded:", decoded)

            assert pdecoded == decoded
            assert pencoded == encoded
        except Exception as e:
            print(f"Error: {e}")


    def test_z85p(s=b"Hello, Z85P!"):
        try:
            pencoded = Z85Ppy.encode(s)
            print(f"Encoded py: {pencoded}")
            pdecoded = Z85Ppy.decode(pencoded)
            print(f"Decoded py: {pdecoded.decode('utf-8')}")

            encoded = Z85P.encode(s)
            print(f"Encoded: {encoded}")
            decoded = Z85P.decode(encoded)
            print(f"Decoded: {decoded.decode('utf-8')}")

            assert pdecoded == decoded
            assert pencoded == encoded
        except Exception as e:
            print(f"Error: {e}")


    test_b91()
    test_z85p()
    test_z85b()



"""
Python implementation of Z85b 85-bit encoding.

Z85b is a variation of ZMQ RFC 32 Z85 85-bit encoding with the following differences:
1. Little-endian encoding (to facilitate alignment with lower byte indices).
2. No requirement for a multiple of 4/5 length.
3. `decode_z85b()` eliminates whitespace from the input.
4. `decode_z85b()` raises a clear exception if invalid characters are encountered.

This file is a derivative work of https://gist.github.com/minrk/6357188?permalink_comment_id=2366506#gistcomment-2366506

Copyright (c) 2013 Brian Granger, Min Ragan-Kelley
Distributed under the terms of the New BSD License.
"""
import re
import struct
from typing import Union


class Z85B:
    # Z85CHARS is the base 85 symbol table
    Z85CHARS = bytearray(b"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.-:+=^!/*?&<>()[]{}@%$#")

    # Z85MAP maps integers in [0, 84] to the appropriate character in Z85CHARS
    Z85MAP = {char: idx for idx, char in enumerate(Z85CHARS)}

    # Powers of 85 for encoding/decoding
    _85s = [85 ** i for i in range(5)]

    # Padding lengths for encoding and decoding
    _E_PADDING = [0, 3, 2, 1]
    _D_PADDING = [0, 4, 3, 2, 1]

    @classmethod
    def encode(cls, data: Union[str, bytes], encoding: str = "utf-8") -> bytes:
        """
        Encode raw bytes into Z85b format.

        Args:
            data (Union[str, bytes]): Input data to encode.
            encoding (str): The encoding to use if `data` is provided as a string. Default is 'utf-8'.

        Returns:
            bytes: Z85b-encoded bytes.
        """
        if isinstance(data, str):
            data = data.encode(encoding)
        data = bytearray(data)
        padding = cls._E_PADDING[len(data) % 4]
        data += b'\x00' * padding
        nvalues = len(data) // 4

        # Pack the raw bytes into little-endian 32-bit integers
        values = struct.unpack(f'<{nvalues}I', data)
        encoded = bytearray()

        for value in values:
            for offset in cls._85s:
                encoded.append(cls.Z85CHARS[(value // offset) % 85])

        # Remove padding characters from the encoded output
        if padding:
            encoded = encoded[:-padding]
        return bytes(encoded)

    @classmethod
    def decode(cls, encoded_data: Union[str, bytes], encoding: str = "utf-8") -> bytes:
        """
        Decode Z85b-encoded bytes into raw bytes.

        Args:
            encoded_data (Union[str, bytes]): Z85b-encoded data.
            encoding (str): The encoding to use if `encoded_data` is provided as a string. Default is 'utf-8'.

        Returns:
            bytes: Decoded raw bytes.

        Raises:
            Z85DecodeError: If invalid characters are encountered during decoding.
        """
        # Normalize input by removing whitespace
        encoded_data = bytearray(re.sub(rb'\s+', b'',
                                        encoded_data if isinstance(encoded_data, bytes)
                                        else encoded_data.encode(encoding)))
        padding = cls._D_PADDING[len(encoded_data) % 5]
        nvalues = (len(encoded_data) + padding) // 5

        values = []
        for i in range(0, len(encoded_data), 5):
            value = 0
            for j, offset in enumerate(cls._85s):
                try:
                    value += cls.Z85MAP[encoded_data[i + j]] * offset
                except IndexError:
                    break  # End of input reached
                except KeyError as e:
                    raise ValueError(f"Invalid byte code: {e.args[0]!r}")
            values.append(value)

        # Unpack the values back into raw bytes
        decoded = struct.pack(f'<{nvalues}I', *values)

        # Remove padding from the decoded output
        if padding:
            decoded = decoded[:-padding]
        return decoded



from typing import Union
import struct

class Z85P:
    """
    Z85 is a class that provides encoding and decoding methods for transforming raw bytes into the Z85 encoding format.
    Z85 encoding represents 32-bit chunks of input bytes into a base85-encoded string with padding applied.
    The padding is added to ensure the encoded data's length is a multiple of 4 characters.
    The first byte of the encoded data indicates how many padding characters were added, which can be removed during decoding.
    """
    Z85CHARS = b"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.-:+=^!/*?&<>()[]{}@%$#"
    Z85MAP = {c: idx for idx, c in enumerate(Z85CHARS)}

    _85s = [85 ** i for i in range(5)][::-1]

    @classmethod
    def encode(cls, rawbytes: Union[str, bytes]) -> bytes:
        """
        Encodes raw bytes into Z85 encoding format with padding, and prepends the padding size.

        Args:
            rawbytes (Union[str, bytes]): The input raw bytes to be encoded.

        Returns:
            bytes: The Z85-encoded byte sequence with appropriate padding and padding size indication.

        Notes:
            The padding is applied to ensure the length of the encoded data is a multiple of 5. The first byte in the
            returned byte sequence represents the number of padding characters added.
        """
        if isinstance(rawbytes, str):
            rawbytes = rawbytes.encode("utf-8")

        padding = (4 - len(rawbytes) % 4) % 4  # Padding to make the length a multiple of 4
        rawbytes += b'\x00' * padding

        # The first byte indicates how many padding characters were added
        nvalues = len(rawbytes) // 4
        values = struct.unpack('>%dI' % nvalues, rawbytes)
        encoded = [padding]

        for v in values:
            for offset in cls._85s:
                encoded.append(cls.Z85CHARS[(v // offset) % 85])

        return bytes(encoded)

    @classmethod
    def decode(cls, z85bytes: Union[str, bytes]) -> bytes:
        """
        Decodes a Z85-encoded byte sequence back into raw bytes, removing padding as indicated by the first byte.

        Args:
            z85bytes (Union[str, bytes]): The Z85-encoded byte sequence to be decoded.

        Returns:
            bytes: The decoded raw byte sequence with padding removed.

        Raises:
            ValueError: If the length of the input data is not divisible by 5 or contains invalid Z85 encoding.

        Notes:
            The first byte of the encoded data indicates the padding size, and this padding is removed during decoding.
        """
        if isinstance(z85bytes, str):
            z85bytes = z85bytes.encode("utf-8")

        if len(z85bytes) == 0:
            return z85bytes

        if len(z85bytes) % 5 != 1:
            raise ValueError('Invalid data length, should be divisible by 5 with 1 extra byte for padding indicator.')

        padding = z85bytes[0]  # Read the padding size from the first byte
        if padding < 0 or padding > 4:
            raise ValueError('Padding size must be between 0 and 4.')

        z85bytes = z85bytes[1:]  # Remove the first byte (padding size byte)

        values = []
        for i in range(0, len(z85bytes), 5):
            value = 0
            for j, offset in enumerate(cls._85s):
                value += cls.Z85MAP[z85bytes[i + j]] * offset
            values.append(value)

        decoded = struct.pack('>%dI' % len(values), *values)
        return decoded[:-padding] if padding else decoded  # Remove padding



from typing import Union


class B91:
    ALPHABET = [
        'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
        'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
        'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
        'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '!', '#', '$',
        '%', '&', '(', ')', '*', '+', ',', '.', '/', ':', ';', '<', '=',
        '>', '?', '@', '[', ']', '^', '_', '`', '{', '|', '}', '~', '"'
    ]

    DECODE_TABLE = {char: idx for idx, char in enumerate(ALPHABET)}

    @classmethod
    def decode(cls, encoded_data: Union[str, bytes], encoding: str = "utf-8") -> bytes:
        """
        Decodes a Base91-encoded string into its original binary form.

        Args:
            encoded_data (Union[str, bytes]): Base91-encoded input data. If `bytes`, it is decoded as UTF-8.
            encoding (str): The encoding to use if `encoded_data` is provided as a string. Default is 'utf-8'.

        Returns:
            bytes: The decoded binary data.

        Raises:
            ValueError: If the input contains invalid Base91 characters.
        """
        if isinstance(encoded_data, bytes):
            encoded_data = encoded_data.decode(encoding)

        v = -1
        b = 0
        n = 0
        out = bytearray()

        for char in encoded_data:
            if char not in cls.DECODE_TABLE:
                raise ValueError(f"Invalid Base91 character: {char}")
            c = cls.DECODE_TABLE[char]
            if v < 0:
                v = c
            else:
                v += c * 91
                b |= v << n
                n += 13 if (v & 8191) > 88 else 14
                while n >= 8:
                    out.append(b & 255)
                    b >>= 8
                    n -= 8
                v = -1

        if v >= 0:
            out.append((b | v << n) & 255)

        return bytes(out)

    @classmethod
    def encode(cls, data: Union[bytes, str], encoding: str = "utf-8") -> bytes:
        """
        Encodes binary data into a Base91-encoded string.

        Args:
            data (Union[bytes, str]): Input binary data to encode. If `str`, it is encoded as UTF-8.
            encoding (str): The encoding to use if `data` is provided as a string. Default is 'utf-8'.

        Returns:
            str: The Base91-encoded string.
        """
        if isinstance(data, str):
            data = data.encode(encoding)

        b = 0
        n = 0
        out = []

        for byte in data:
            b |= byte << n
            n += 8
            if n > 13:
                v = b & 8191
                if v > 88:
                    b >>= 13
                    n -= 13
                else:
                    v = b & 16383
                    b >>= 14
                    n -= 14
                out.append(cls.ALPHABET[v % 91])
                out.append(cls.ALPHABET[v // 91])

        if n:
            out.append(cls.ALPHABET[b % 91])
            if n > 7 or b > 90:
                out.append(cls.ALPHABET[b // 91])

        return ''.join(out).encode(encoding)



import unittest
from z85base91 import Z85B


class TestZ85B(unittest.TestCase):
    def test_encode_empty(self):
        """Test encoding an empty byte sequence."""
        self.assertEqual(Z85B.encode(b''), b'')
        self.assertEqual(Z85B.encode(''), b'')

    def test_decode_empty(self):
        """Test decoding an empty string."""
        self.assertEqual(Z85B.decode(''), b'')
        self.assertEqual(Z85B.decode(b''), b'')

    def test_encode_single_byte(self):
        """Test encoding a single byte."""
        self.assertEqual(b'A', Z85B.decode(Z85B.encode(b'A')))
        self.assertEqual(b'B', Z85B.decode(Z85B.encode('B')))
        self.assertEqual(b'_~', Z85B.decode(Z85B.encode(b'_~')))
        self.assertEqual(b'_~', Z85B.decode(Z85B.encode('_~')))

    def test_encode_short_string(self):
        """Test encoding a short string."""
        self.assertEqual(b'hello', Z85B.decode(Z85B.encode(b'hello')))
        self.assertEqual(b'Hello World', Z85B.decode(Z85B.encode(b'Hello World')))

    def test_encode_decode_round_trip(self):
        """Test encoding and decoding round-trip."""
        data = b'The quick brown fox jumps over the lazy dog.'
        encoded = Z85B.encode(data)
        decoded = Z85B.decode(encoded)
        self.assertEqual(decoded, data)

    def test_encode_unicode_string(self):
        """Test encoding a Unicode string."""
        data = 'こんにちは'  # Japanese for "hello"
        encoded = Z85B.encode(data)
        decoded = Z85B.decode(encoded)
        self.assertEqual(decoded.decode('utf-8'), data)

    def test_decode_invalid_character(self):
        """Test decoding with invalid Base91 characters."""
        with self.assertRaises(ValueError):
            Z85B.decode('Invalid🎉Chars')

    def test_edge_case_88_threshold(self):
        """Test edge cases around the 88 threshold."""
        data = b'\x00\x00\x00'  # Minimal data
        encoded = Z85B.encode(data)
        self.assertEqual(Z85B.decode(encoded), data)

    def test_encode_large_data(self):
        """Test encoding a large byte sequence."""
        data = b'\xff' * 1000
        encoded = Z85B.encode(data)
        decoded = Z85B.decode(encoded)
        self.assertEqual(decoded, data)


if __name__ == '__main__':
    unittest.main()



import os
from setuptools import setup

BASEDIR = os.path.abspath(os.path.dirname(__file__))


def get_version():
    """ Find the version of the package"""
    version_file = os.path.join(BASEDIR, 'z85base91', 'version.py')
    major, minor, build, alpha = (None, None, None, None)
    with open(version_file) as f:
        for line in f:
            if 'VERSION_MAJOR' in line:
                major = line.split('=')[1].strip()
            elif 'VERSION_MINOR' in line:
                minor = line.split('=')[1].strip()
            elif 'VERSION_BUILD' in line:
                build = line.split('=')[1].strip()
            elif 'VERSION_ALPHA' in line:
                alpha = line.split('=')[1].strip()

            if ((major and minor and build and alpha) or
                    '# END_VERSION_BLOCK' in line):
                break
    version = f"{major}.{minor}.{build}"
    if int(alpha):
        version += f"a{alpha}"
    return version


def required(requirements_file):
    """ Read requirements file and remove comments and empty lines. """
    with open(os.path.join(BASEDIR, requirements_file), 'r') as f:
        requirements = f.read().splitlines()
        return [pkg for pkg in requirements
                if pkg.strip() and not pkg.startswith("#")]



setup(
    name='z85base91',
    version=get_version(),
    packages=['z85base91'],
    url='https://github.com/JarbasHiveMind/z85base91',
    license='Apache-2.0',
    author='jarbasAi',
    include_package_data=True,
    author_email='jarbasai@mailfence.com',
    description='base91, z85b and z85p encodings'
)



import unittest
from z85base91 import Z85P


class TestZ85P(unittest.TestCase):
    def test_encode_empty(self):
        """Test encoding an empty byte sequence."""
        self.assertEqual(Z85P.encode(b''), b'\x00')
        self.assertEqual(Z85P.encode(''), b'\x00')

    def test_decode_empty(self):
        """Test decoding an empty string."""
        self.assertEqual(Z85P.decode(b'\x00'), b'')
        self.assertEqual(Z85P.decode(''), b'')
        self.assertEqual(Z85P.decode(b''), b'')

    def test_encode_single_byte(self):
        """Test encoding a single byte."""
        self.assertEqual(b'A', Z85P.decode(Z85P.encode(b'A')))
        self.assertEqual(b'B', Z85P.decode(Z85P.encode('B')))
        self.assertEqual(b'_~', Z85P.decode(Z85P.encode(b'_~')))
        self.assertEqual(b'_~', Z85P.decode(Z85P.encode('_~')))

    def test_encode_short_string(self):
        """Test encoding a short string."""
        self.assertEqual(b'hello', Z85P.decode(Z85P.encode(b'hello')))

    def test_encode_decode_round_trip(self):
        """Test encoding and decoding round-trip."""
        data = b'The quick brown fox jumps over the lazy dog.'
        encoded = Z85P.encode(data)
        decoded = Z85P.decode(encoded)
        self.assertEqual(decoded, data)

    def test_encode_unicode_string(self):
        """Test encoding a Unicode string."""
        data = 'こんにちは'  # Japanese for "hello"
        encoded = Z85P.encode(data)
        decoded = Z85P.decode(encoded)
        self.assertEqual(decoded.decode('utf-8'), data)

    def test_decode_invalid_character(self):
        """Test decoding with invalid z85 characters."""
        with self.assertRaises(ValueError):
            Z85P.decode('Invalid🎉Chars')

    def test_encode_large_data(self):
        """Test encoding a large byte sequence."""
        data = b'\xff' * 1000
        encoded = Z85P.encode(data)
        decoded = Z85P.decode(encoded)
        self.assertEqual(decoded, data)

    def test_padding_single_byte(self):
        """Test encoding and decoding with one byte that requires padding."""
        data = b'\x01'  # Single byte, should get padded with 3 \x00 bytes
        encoded = Z85P.encode(data)
        self.assertEqual(encoded[0], 3)  # Check padding byte
        self.assertEqual(Z85P.decode(encoded), data)

    def test_padding_two_bytes(self):
        """Test encoding and decoding with two bytes that require padding."""
        data = b'\x01\x01'  # Two bytes, should get padded with 2 \x00 bytes
        encoded = Z85P.encode(data)
        self.assertEqual(encoded[0], 2)  # Check padding byte
        self.assertEqual(Z85P.decode(encoded), data)

    def test_padding_three_bytes(self):
        """Test encoding and decoding with three bytes that require padding."""
        data = b'\x01\x01\x01'  # Three bytes, should get padded with 1 \x00 byte
        encoded = Z85P.encode(data)
        self.assertEqual(encoded[0], 1)  # Check padding byte
        self.assertEqual(Z85P.decode(encoded), data)

    def test_no_padding_5_bytes(self): # fails
        data = b'\x01\x01\x01\x01\x01'  # 5 bytes
        encoded = Z85P.encode(data)
        self.assertEqual(encoded[0], 3)  # Check padding byte
        self.assertEqual(Z85P.decode(encoded), data)

    def test_no_padding_needed(self):
        """Test encoding and decoding with data that doesn't need padding."""
        data = b'\x01\x01\x01\x01'  # Exactly 4 bytes, no padding
        encoded = Z85P.encode(data)
        self.assertEqual(encoded[0], 0)  # No padding
        self.assertEqual(Z85P.decode(encoded), data)

    def test_round_trip_padding(self):
        """Test round-trip encoding and decoding with padding."""
        data = b'\x01\x01\x01'  # Less than 4 bytes, needs padding
        encoded = Z85P.encode(data)
        decoded = Z85P.decode(encoded)
        self.assertEqual(decoded, data)  # Ensure padding is correctly removed

    def test_padding_removal_after_decoding(self):
        """Test ensuring padding is correctly removed after decoding."""
        data = b'\x01\x01\x01'  # Less than 4 bytes, needs padding
        encoded = Z85P.encode(data)
        self.assertEqual(encoded[0], 1)  # Padding size is 1
        decoded = Z85P.decode(encoded)
        self.assertEqual(decoded, data)  # Padding should be removed

if __name__ == '__main__':
    unittest.main()
